import csv

class QuerySetCSVExporter:

    def __init__(self, queryset, filename):
        self.queryset = queryset
        self.filename = filename

    def export(self):
        fields_dicts = self.queryset.values()
        if len(fields_dicts) == 0:
            return
        with open(self.filename, 'w', newline='', encoding='utf-8') as csvfile:
            writer = csv.writer(csvfile, delimiter=',')
            writer.writerow(fields_dicts[0].keys())
            for fields_dict in fields_dicts:
                writer.writerow(fields_dict.values())



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _

class DjangoOldRecordsConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'django_old_records'
    verbose_name = _("Old Records")



from django.contrib import admin
from django_old_records.models import ModelConfig, FieldConfig

class FieldConfigInline(admin.TabularInline):
    model = FieldConfig

class ModelConfigAdmin(admin.ModelAdmin):
    inlines = [FieldConfigInline]

    def get_queryset(self, request):
        return super().get_queryset(request).select_related('content_type')

admin.site.register(ModelConfig, ModelConfigAdmin)



from django.apps import apps
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Q

def validate_field(value):
    print(value, type(value))

def validate_field_value(value):
    print(value, type(value))

def limit_choices_to():
    query_filter = Q()
    for key in getattr(settings, 'OLD_RECORDS_MODELS', {}):
        for model_name in settings.OLD_RECORDS_MODELS[key]:
            model = apps.get_model(app_label=key, model_name=model_name)
            query_filter |= Q(app_label=model._meta.app_label, model=model._meta.model_name)
    return query_filter

class ModelConfig(models.Model):

    content_type = models.OneToOneField(
        ContentType,
        limit_choices_to=limit_choices_to,
        on_delete=models.CASCADE,
    )
    max_age = models.DurationField(null=True, blank=True)
    export = models.BooleanField(default=False)
    created_at_field = models.CharField(max_length=100, null=True, blank=True)

    def __str__(self):
        return self.content_type.model_class().__name__

class FieldConfig(models.Model):

    model_config = models.ForeignKey('ModelConfig', on_delete=models.CASCADE)
    field_name = models.CharField(max_length=100, validators=[validate_field])
    value = models.CharField(max_length=100, validators=[validate_field_value])
    max_age = models.DurationField()



# Generated by Django 3.2.4 on 2021-09-24 20:19

from django.db import migrations, models
import django.db.models.deletion
import django_old_records.models


class Migration(migrations.Migration):

    replaces = [('django_old_records', '0001_initial'), ('django_old_records', '0002_modelconfig_mode'), ('django_old_records', '0003_alter_modelconfig_content_type'), ('django_old_records', '0004_auto_20210921_1928'), ('django_old_records', '0005_alter_modelconfig_content_type'), ('django_old_records', '0006_alter_modelconfig_max_age')]

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='ModelConfig',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('max_age', models.DurationField(blank=True, null=True)),
                ('content_type', models.OneToOneField(limit_choices_to=django_old_records.models.limit_choices_to, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
                ('export', models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name='FieldConfig',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('field_name', models.CharField(max_length=100, validators=[django_old_records.models.validate_field])),
                ('value', models.CharField(max_length=100, validators=[django_old_records.models.validate_field_value])),
                ('max_age', models.DurationField()),
                ('model_config', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_old_records.modelconfig')),
            ],
        ),
    ]



# Generated by Django 3.2.4 on 2021-09-28 13:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_old_records', '0001_initial_squashed_0006_alter_modelconfig_max_age'),
    ]

    operations = [
        migrations.AddField(
            model_name='modelconfig',
            name='created_at_field',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
    ]



from django.db import models
from datetime import timedelta
from django.conf import settings
from django.utils import timezone
from django.db.models import Q
from django.apps import apps

class OldRecordsManager(models.Manager):
    DEFAULT_CREATED_AT_FIELD = getattr(settings, 'OLD_RECORDS_DEFAULT_CREATED_AT_FIELD', 'created_at')
    DEFAULT_MAX_AGE = getattr(settings, 'OLD_RECORDS_DEFAULT_MAX_AGE', None)

    def _initialize(self):
        self.model_config = self._get_model_config()

    def _created_at_field(self):
        return self._model_config_created_at_field() or getattr(self.model, 'created_at_field', self.DEFAULT_CREATED_AT_FIELD)

    def _get_model_config(self):
        from .models import ModelConfig
        from django.contrib.contenttypes.models import ContentType

        content_type = ContentType.objects.get_for_model(self.model)
        return ModelConfig.objects.filter(content_type=content_type).prefetch_related('fieldconfig_set').first()

    def _model_config_created_at_field(self):
        return self.model_config.created_at_field if self.model_config else None

    def _model_config_max_age(self):
        return self.model_config.max_age if self.model_config else None

    def _model_config_field_configs(self):
        model = apps.get_model(app_label='django_old_records', model_name='FieldConfig')
        return self.model_config.fieldconfig_set.all() if self.model_config else model.objects.none()

    def _default_max_age(self):
        max_age = getattr(self.model, 'max_age', self.DEFAULT_MAX_AGE)
        if isinstance(max_age, int):
            return timedelta(days=max_age)

        if isinstance(max_age, timedelta):
            return max_age

    def created_at_lookup(self):
        return f'{self._created_at_field()}__lte'

    def _field_config_queries(self, field_config):
        time_limit = timezone.now() - field_config.max_age
        return Q(**{field_config.field_name: field_config.value}), Q(**{self.created_at_lookup(): time_limit})

    def _build_filter_query(self):
        model_max_age = self._model_config_max_age() if self._model_config_max_age() else self._default_max_age()
        model_time_limit =  timezone.now() - model_max_age if model_max_age else None
        field_configs = self._model_config_field_configs()
        field_filter_query = None
        model_filter_query = None

        if not field_configs and model_time_limit:
            return Q(**{self.created_at_lookup(): model_time_limit})

        if field_configs:
            value_query, age_query = self._field_config_queries(field_configs[0])
            field_filter_query = value_query & age_query
            model_filter_query = ~value_query

            for field_config in field_configs[1:]:
                value_query, age_query = self._field_config_queries(field_config)
                field_filter_query |= value_query & age_query
                model_filter_query &= ~value_query

        if model_time_limit and model_filter_query:
            model_filter_query &= Q(**{self.created_at_lookup(): model_time_limit})

        if model_time_limit and field_filter_query:
            return model_filter_query | field_filter_query
        return field_filter_query

    def get_queryset(self):
        self._initialize()
        if not hasattr(self.model, self._created_at_field()):
            return super().get_queryset().none()


        query = self._build_filter_query()
        if query:
            return super().get_queryset().filter(query)
        return super().get_queryset().none()



from django.core.management.base import BaseCommand
from django_old_records.old_records_manager import OldRecordsManager
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.apps import apps
from django_old_records.models import ModelConfig
from django_old_records.queryset_csv_exporter import QuerySetCSVExporter

class Command(BaseCommand):
    help = u"""Deletes all old records from database"""

    def handle(self, *args, **kwargs):
        content_type_filter_query = Q(pk__in=[])

        for app in getattr(settings, 'OLD_RECORDS_MODELS', {}):
            for model_name in settings.OLD_RECORDS_MODELS[app]:
                model = apps.get_model(app_label=app, model_name=model_name)
                content_type_filter_query |= Q(app_label=app, model=model._meta.model_name)
        content_types = ContentType.objects.filter(content_type_filter_query)

        for content_type in content_types:
            model_config = ModelConfig.objects.filter(content_type=content_type).first()
            model = content_type.model_class()
            if hasattr(model, 'old_records') and isinstance(model.old_records, OldRecordsManager):
                self.stdout.write(f'Deleting {model.old_records.count()} {model._meta.app_label}.{model.__name__} old records')
                if model_config and model_config.export:
                    filename = f'{model._meta.app_label}_{model._meta.model_name}_old_records.csv'
                    QuerySetCSVExporter(model.old_records.all(), filename).export()
                    self.stdout.write(f'Exported deleted records to {filename}')
                model.old_records.all().delete()
